﻿/*
 * Created by SharpDevelop.
 * User: Ouroboros
 * Date: 2006.11.04.
 * Time: 14:55
 * Version: 0.3.0
 * Latest update: property calls replaced with field calls to improve performance.
 * Updates:
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Runtime.InteropServices;

namespace TetheredSun.Mathematics
{
	/// <summary>
	/// A class representing complex numbers.
	/// A modified copy of System.Numerics.Complex in .NET 4.
	/// </summary>
	[Serializable, StructLayout(LayoutKind.Sequential)]
	public struct Complex : IEquatable<Complex>, IFormattable
	{
	    private double real;
	    private double imaginary;
	    public static readonly Complex Zero;
	    public static readonly Complex One;
	    public static readonly Complex ImaginaryOne;
	    private const double log10Inv = 0.43429448190325;
	    
	    static Complex()
	    {
	        Zero = new Complex(0.0, 0.0);
	        One = new Complex(1.0, 0.0);
	        ImaginaryOne = new Complex(0.0, 1.0);
	    }
	    
	    public Complex(double real, double imaginary)
	    {
	        this.real = real;
	        this.imaginary = imaginary;
	    }
	    
	    public double Real {
	        get { return this.real; }
	    }
	    
	    public double Imaginary {
			get { return this.imaginary; }
	    }
	    
	    public double Magnitude {
	    	get { return Abs(this); }
	    }
	    
	    public double Phase {
	        get { return Math.Atan2(this.imaginary, this.real); }
	    }

	    
	    public static Complex FromPolarCoordinates(double magnitude, double phase)
	    {
	        return new Complex(magnitude * Math.Cos(phase), magnitude * Math.Sin(phase));
	    }
	    
	    public static Complex Negate(Complex value)
	    {
	        return -value;
	    }
	
	 
	    public static Complex Add(Complex left, Complex right)
	    {
	        return left + right;
	    }
	
	
	    public static Complex Subtract(Complex left, Complex right)
	    {
	        return left - right;
	    }
	
	 
	    public static Complex Multiply(Complex left, Complex right)
	    {
	        return left * right;
	    }
	
	  
	    public static Complex Divide(Complex dividend, Complex divisor)
	    {
	        return dividend / divisor;
	    }
	
	    public static Complex operator -(Complex value)
	    {
	        return new Complex(-value.real, -value.imaginary);
	    }
	
	    public static Complex operator +(Complex left, Complex right)
	    {
	        return new Complex(left.real + right.real, left.imaginary + right.imaginary);
	    }
	
	    public static Complex operator -(Complex left, Complex right)
	    {
	        return new Complex(left.real - right.real, left.imaginary - right.imaginary);
	    }
	
	    public static Complex operator *(Complex left, Complex right)
	    {
	        double real = (left.real * right.real) - (left.imaginary * right.imaginary);
	        return new Complex(real, (left.imaginary * right.real) + (left.real * right.imaginary));
	    }
	
	    public static Complex operator /(Complex left, Complex right)
	    {
	        double real = left.real;
	        double imaginary = left.imaginary;
	        double num3 = right.real;
	        double num4 = right.imaginary;
	        
	        if (Math.Abs(num4) < Math.Abs(num3)) {
	            return new Complex((real + (imaginary * (num4 / num3))) / (num3 + (num4 * (num4 / num3))), (imaginary - (real * (num4 / num3))) / (num3 + (num4 * (num4 / num3))));
	        }
	        
	        return new Complex((imaginary + (real * (num3 / num4))) / (num4 + (num3 * (num3 / num4))), (-real + (imaginary * (num3 / num4))) / (num4 + (num3 * (num3 / num4))));
	    }
	
	    public static double Abs(Complex value)
	    {
	        if (double.IsInfinity(value.real) || double.IsInfinity(value.imaginary)) {
	            return double.PositiveInfinity;
	        }
	    	
	        double num = Math.Abs(value.real);
	        double num2 = Math.Abs(value.imaginary);
	        
	        if (num > num2) {
	            double num3 = num2 / num;
	            return (num * Math.Sqrt(1.0 + (num3 * num3)));
	        }
	        
	        if (num2 == 0.0) {
	            return num;
	        }
	        double num4 = num / num2;
	        
	        return num2 * Math.Sqrt(1.0 + (num4 * num4));
	    }
	
	    public static Complex Conjugate(Complex value)
	    {
	        return new Complex(value.real, -value.imaginary);
	    }
	
	    public static Complex Reciprocal(Complex value)
	    {
	        if ((value.real == 0.0) && (value.imaginary == 0.0))
	        {
	            return Zero;
	        }
	        return (One / value);
	    }
	
	    public static bool operator ==(Complex left, Complex right)
	    {
	        return (left.real == right.real) && (left.imaginary == right.imaginary);
	    }
	
	    public static bool operator !=(Complex left, Complex right)
	    {
	        if (left.real == right.real) {
	            return left.imaginary != right.imaginary;
	        }
	        return true;
	    }
	
	    public override bool Equals(object obj)
	    {
	        return (obj is Complex) && (this == ((Complex)obj));
	    }
	
	    public bool Equals(Complex value)
	    {
	        return this.real.Equals(value.real) && this.imaginary.Equals(value.imaginary);
	    }
	
	    public static implicit operator Complex(short value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(int value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(long value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(ushort value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(uint value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(ulong value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(sbyte value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(byte value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(float value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public static implicit operator Complex(double value)
	    {
	        return new Complex(value, 0.0);
	    }
	
	    public static explicit operator Complex(decimal value)
	    {
	        return new Complex((double) value, 0.0);
	    }
	
	    public override string ToString()
	    {
	        return string.Format(CultureInfo.CurrentCulture, "({0}, {1})", new object[] { this.real, this.imaginary });
	    }
	
	    public string ToString(string format)
	    {
	        return string.Format(CultureInfo.CurrentCulture, "({0}, {1})", new object[] { this.real.ToString(format, CultureInfo.CurrentCulture), this.imaginary.ToString(format, CultureInfo.CurrentCulture) });
	    }
	
	    public string ToString(IFormatProvider provider)
	    {
	        return string.Format(provider, "({0}, {1})", new object[] { this.real, this.imaginary });
	    }
	
	    public string ToString(string format, IFormatProvider provider)
	    {
	        return string.Format(provider, "({0}, {1})", new object[] { this.real.ToString(format, provider), this.imaginary.ToString(format, provider) });
	    }
	
	    public override int GetHashCode()
	    {
	        int num = 0x5f5e0fd;
	        int num2 = this.real.GetHashCode() % num;
	        int hashCode = this.imaginary.GetHashCode();
	        return num2 ^ hashCode;
	    }
	
	    public static Complex Sin(Complex value)
	    {
	        double real = value.real;
	        double imaginary = value.imaginary;
	        return new Complex(Math.Sin(real) * Math.Cosh(imaginary), Math.Cos(real) * Math.Sinh(imaginary));
	    }
	
	    public static Complex Sinh(Complex value)
	    {
	        double real = value.real;
	        double imaginary = value.imaginary;
	        return new Complex(Math.Sinh(real) * Math.Cos(imaginary), Math.Cosh(real) * Math.Sin(imaginary));
	    }
	
	    public static Complex Asin(Complex value)
	    {
	        return -ImaginaryOne * Log((ImaginaryOne * value) + Sqrt(One - (value * value)));
	    }
	
	    public static Complex Cos(Complex value)
	    {
	        double real = value.real;
	        double imaginary = value.imaginary;
	        return new Complex(Math.Cos(real) * Math.Cosh(imaginary), -(Math.Sin(real) * Math.Sinh(imaginary)));
	    }
	
	    public static Complex Cosh(Complex value)
	    {
	        double real = value.real;
	        double imaginary = value.imaginary;
	        return new Complex(Math.Cosh(real) * Math.Cos(imaginary), Math.Sinh(real) * Math.Sin(imaginary));
	    }
	
	    public static Complex Acos(Complex value)
	    {
	        return -ImaginaryOne * Log(value + (ImaginaryOne * Sqrt(One - (value * value))));
	    }
	
	    public static Complex Tan(Complex value)
	    {
	        return Sin(value) / Cos(value);
	    }
	
	    public static Complex Tanh(Complex value)
	    {
	        return Sinh(value) / Cosh(value);
	    }
	
	    public static Complex Atan(Complex value)
	    {
	        Complex complex = new Complex(2.0, 0.0);
	        return (ImaginaryOne / complex) * (Log(One - (ImaginaryOne * value)) - Log(One + (ImaginaryOne * value)));
	    }
	
	    public static Complex Log(Complex value)
	    {
	        return new Complex(Math.Log(Abs(value)), Math.Atan2(value.imaginary, value.real));
	    }
	
	    public static Complex Log(Complex value, double baseValue)
	    {
	        return Log(value) / Log(baseValue);
	    }
	
	    public static Complex Log10(Complex value)
	    {
	        return Scale(Log(value), 0.43429448190325);
	    }
	
	    public static Complex Exp(Complex value)
	    {
	        double num = Math.Exp(value.real);
	        double real = num * Math.Cos(value.imaginary);
	        return new Complex(real, num * Math.Sin(value.imaginary));
	    }
	
	    public static Complex Sqrt(Complex value)
	    {
	        return FromPolarCoordinates(Math.Sqrt(value.Magnitude), value.Phase / 2.0);
	    }
	
	    public static Complex Pow(Complex value, Complex power)
	    {
	        if (power == Zero) {
	            return One;
	        }
	    	
	        if (value == Zero) {
	            return Zero;
	        }
	    	
	        double real = value.real;
	        double imaginary = value.imaginary;
	        double y = power.real;
	        double num4 = power.imaginary;
	        double d = Abs(value);
	        double num6 = Math.Atan2(imaginary, real);
	        double num7 = (y * num6) + (num4 * Math.Log(d));
	        double num8 = Math.Pow(d, y) * Math.Pow(2.7182818284590451, -num4 * num6);
	        
	        return new Complex(num8 * Math.Cos(num7), num8 * Math.Sin(num7));
	    }
	
	    public static Complex Pow(Complex value, double power)
	    {
	        return Pow(value, new Complex(power, 0.0));
	    }
	
	    private static Complex Scale(Complex value, double factor)
	    {
	        double real = factor * value.real;
	        return new Complex(real, factor * value.imaginary);
	    }
	}
}
