﻿/*
 * Created by SharpDevelop.
 * User: phil
 * Date: 2010.09.29.
 * Time: 14:25
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Windows.Forms;

namespace TetheredSun.Mathematics
{
	/// <summary>
	/// Description of Matrix.
	/// </summary>
	public class Matrix
	{
		#region FIELDS:
		public static double Epsilon = Math.Pow(10.0, -10.0);
		protected double[][] items;
		#endregion
		
		#region CONSTRUCTORS:
		public Matrix(int rowCount, int columnCount) : this(rowCount, columnCount, 0.0, false) { }
		
		public Matrix(double[,] items)
		{
			this.items = new double[items.GetLength(0)][];
			int m = items.GetLength(0);
			int n = items.GetLength(1);
			
			for (int i = 0; i < m; i++) {
				this.items[i] = new double[n];
				for (int j = 0; j < n; j++) {
					this.items[i][j] = items[i, j];
				}
			}
		}
		
		public Matrix(int rowCount, int columnCount, double value) : this(rowCount, columnCount, value, true) { }
		
		protected Matrix(int rowCount, int columnCount, double value, bool initialise)
		{
			items = new double[rowCount][];
			
			for (int i = 0; i < rowCount; i++) {
				items[i] = new double[columnCount];
				if (!initialise) continue;
				for (int j = 0; j < columnCount; j++) {
					items[i][j] = value;
				}
			}
		}
		
		protected Matrix(double[][] items, bool clone)
		{
			if (items == null) {
				throw new NullReferenceException("Trying to construct a matrix from a null array.");
			}
			for (int i = 0; i < items.Length; i++) {
				if (items[i] == null) {
					throw new NullReferenceException("Trying to construct a matrix row from a null array.");
				}
				if (items[i].Length != items[0].Length) {
					Exception exception = new ArgumentException("Row length mismatch.");
					exception.Data.Add("First row length", items[0].Length);
					exception.Data.Add("Actual row", i);
					exception.Data.Add("Actual row length", items[i].Length);
					throw exception;
				}
			}
			if (!clone) {
				this.items = items;
			} else {
				this.items = new double[items.Length][];
				for (int i = 0; i < this.items.Length; i++) {
					this.items[i] = new double[items[0].Length];
					for (int j = 0; j < this.items[i].Length; j++) {
						this.items[i][j] = items[i][j];
					}
				}
			}
		}
		
		public Matrix(double[][] items) : this(items, true) { }
		#endregion		
		
		#region PROPERTIES:
		public int ColumnCount {
			get { return items[0].Length; }
		}
		
		public int RowCount {
			get { return items.Length; }
		}
		
		public double[] this[int index] {
			get { return items[index]; }
		}
		
		public double this[int rowIndex, int columnIndex] {
			get { return items[rowIndex][columnIndex]; }
			set { items[rowIndex][columnIndex] = value; }
		}
		
		public Matrix Inverse {
			get {
				if (items.Length != items[0].Length) {
					Exception exception = new InvalidOperationException(
						"Non-square matrices have no inverse.");
					exception.Data.Add("Rows", items.Length);
					exception.Data.Add("Columns", items[0].Length);
					throw exception;
				}
				return Solve(items, CreateIdentityArray(items.Length));
			}
		}
		
		public Matrix Transpose {
			get {
				Matrix transpose = new Matrix(items[0].Length, items.Length);
				
				for (int i = 0; i < items[0].Length; i++) {
					for (int j = 0; j < items.Length; j++) {
						transpose.items[i][j] = items[j][i];
					}
				}
				
				return transpose;
			}
		}
		#endregion
		
		#region METHODS:
		public static double[][] Add(double[][] matrix1, double[][] matrix2)
		{
			Exception exception;
			int rows;
			int columns;
			double[][] result;
			
			if (CheckAdditiveCompatibility(matrix1, matrix2, out rows, out columns, out exception)) {
				result = new double[rows][];
				
				for (int i = 0; i < rows; i++) {
					result[i] = new double[columns];
					for (int j = 0; j < columns; j++) {
						result[i][j] = matrix1[i][j] + matrix2[i][j];
					}
				}
				
				return result;
			} else {
				throw exception;
			}
		}
		
		public static Matrix Add(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Add(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix Add(double[][] matrix1, Matrix matrix2)
		{
			return new Matrix(Add(matrix1, matrix2.items), false);
		}
		
		public static Matrix Add(Matrix matrix1, double[][] matrix2)
		{
			return new Matrix(Add(matrix1.items, matrix2), false);
		}
		
		protected static bool CheckAdditiveCompatibility(double[][] matrix1, double[][] matrix2, out int rows, out int columns, out Exception exception)
		{
			int m1 = matrix1.Length;
			int m2 = matrix2.Length;
			int n1 = matrix1[0].Length;
			int n2 = matrix2[0].Length;
			bool compatible = (m1 == m2) && (n1 == n2);
			
			if (compatible) {
				rows = m1;
				columns = n1;
				exception = null;
			} else {
				rows = -1;
				columns = -1;
				exception = new InvalidOperationException("The dimensions of the matrix operands in addition or subtraction do not match.");
				exception.Data.Add("Rows of matrix1", matrix1.Length);
				exception.Data.Add("Rows of matrix2", matrix2.Length);
				exception.Data.Add("Columns of matrix1", matrix1[0].Length);
				exception.Data.Add("Columns of matrix2", matrix2[0].Length);
			}
			
			return compatible;
		}
		
		public Matrix Clone()
		{
			return new Matrix(this.items, true);
		}
		
		protected static double[][] CreateAugmentedArray(double[][] left, double[][] right)
		{
			if (left.Length != right.Length) return null;
			
			double[][] augmented = new double[left.Length][];
			int columns = left[0].Length + right[0].Length;
			
			for (int i = 0; i < augmented.Length; i++) {
				augmented[i] = new double[columns];
				for (int j = 0; j < left[0].Length; j++) {
					augmented[i][j] = left[i][j];
				}
				for (int j = 0; j < right[0].Length; j++) {
					augmented[i][left[0].Length + j] = right[i][j];
				}
			}
			
			return augmented;
		}
		
		public static Matrix CreateColumnVector(double[] vector)
		{
			Matrix columnVector = new Matrix(vector.Length, 1);
			
			for (int i = 0; i < vector.Length; i++) {
				columnVector[i][0] = vector[i];
			}
			
			return columnVector;
		}
		
		public static Matrix CreateDiagonal(double[] diagonalValues)
		{
			int n = diagonalValues.Length;
			Matrix result = new Matrix(n, n, 0.0);
			
			for (int i = 0; i < n; i++) {
				result.items[i][i] = diagonalValues[i];
			}
			
			return result;
		}
		
		protected static double[][] CreateIdentityArray(int dimension)
		{
			if (dimension < 1) return null;
			
			double[][] identity = new double[dimension][];
			
			for (int i = 0; i < dimension; i++) {
				identity[i] = new double[dimension];
				identity[i][i] = 1.0;
			}
			
			return identity;
		}
		
		public static Matrix CreateIdentityMatrix(int dimension)
		{
			if (dimension < 1) return null;
			
			Matrix identity = new Matrix(dimension, dimension);
			
			for (int i = 0; i < dimension; i++) {
				identity[i][i] = 1.0;
			}
			
			return identity;
		}
		
		public static Matrix CreateRandom(int rowCount, int columnCount)
		{
			Matrix result = new Matrix(rowCount, columnCount);
			Random random = new Random();
			
			for (int i = 0; i < rowCount; i++) {
				for (int j = 0; j < columnCount; j++) {
					result.items[i][j] = random.NextDouble();
				}
			}
			
			return result;
		}
		
		/*def gauss_jordan(m, eps = 1.0/(10**10)):
		  """Puts given matrix (2D array) into the Reduced Row Echelon Form.
		     Returns True if successful, False if 'm' is singular.
		     NOTE: make sure all the matrix items support fractions! Int matrix will NOT work!
		     Written by Jarno Elonen in April 2005, released into Public Domain"""
		  (h, w) = (len(m), len(m[0])
		  for y in range(0,h):
		    maxrow = y
		    for y2 in range(y+1, h):    # Find max pivot
		      if abs(m[y2][y]) > abs(m[maxrow][y]):
		        maxrow = y2
		    (m[y], m[maxrow]) = (m[maxrow], m[y])
		    if abs(m[y][y]) <= eps:     # Singular?
		      return False
		    for y2 in range(y+1, h):    # Eliminate column y
		      c = m[y2][y] / m[y][y]
		      for x in range(y, w):
		        m[y2][x] -= m[y][x] * c
		  for y in range(h-1, 0-1, -1): # Backsubstitute
		    c  = m[y][y]
		    for y2 in range(0,y):
		      for x in range(w-1, y-1, -1):
		        m[y2][x] -=  m[y][x] * m[y2][y] / c
		    m[y][y] /= c
		    for x in range(h, w):       # Normalize row y
		      m[y][x] /= c
		  return True
		*/
		// Written by Jarno Elonen in April 2005, released into Public Domain
		protected static bool GaussJordan(ref double[][] augmentedArray)
		{
			int rows = augmentedArray.Length;
			int columns = augmentedArray[0].Length;
			int maxRow;
			double c;
			
			for (int i = 0; i < rows; i++) {
				maxRow = i;
				for (int i2 = i + 1; i2 < rows; i2++) { // Find max pivot.
					if (Math.Abs(augmentedArray[i2][i]) > Math.Abs(augmentedArray[maxRow][i])) {
						maxRow = i2;
					}
				}
				
				SwapRows(ref augmentedArray, i, maxRow);
				
				if (Math.Abs(augmentedArray[i][i]) < Epsilon) {
					// The matrix is singular:
					return false;
				}
				
				for (int i2 = i + 1; i2 < rows; i2++) { // Eliminate column i.
					c = augmentedArray[i2][i] / augmentedArray[i][i];
					for (int j = i; j < columns; j++) {
						augmentedArray[i2][j] -= augmentedArray[i][j] * c;
					}
				}
			}
			
			// Backsubstitution:
			for (int i = rows - 1; i > -1; i--) {
				c = augmentedArray[i][i];
				
				for (int i2 = 0; i2 < i; i2++) {
					for (int j = columns - 1; j > i - 1; j--) {
						augmentedArray[i2][j] -=  augmentedArray[i][j] * augmentedArray[i2][i] / c;
					}
				}
				
				augmentedArray[i][i] /= c;
				
				// Normalise row i:
				for (int j = rows; j < columns; j++) {
					augmentedArray[i][j] /= c;
				}
			}
			
			return true;
		}
		
		public double[] GetColumn(int index)
		{
			int m = items.Length;
			double[] column = new double[m];
			
			for (int i = 0; i < m; i++) {
				column[i] = items[i][index];
			}
			
			return column;
		}
	
		public double[] GetRow(int rowIndex)
		{
			int n = items[rowIndex].Length;
			double[] rowClone = new double[n];
			
			for (int j = 0; j < n; j++) {
				rowClone[j] = items[rowIndex][j];
			}
			
			return rowClone;
		}
		
		public static double[][] Multiply(double[][] matrix1, double[][] matrix2)
		{
			if (matrix1[0].Length != matrix2.Length) {
				Exception exception = new ArgumentException(
					"Column count of operand 1 must equal row count of operand 2 in matrix product.");
				exception.Data.Add("Column count of operand 1", matrix1[0].Length);
				exception.Data.Add("Row count of operand 2", matrix2.Length);
				throw exception;
			}
			
			int m = matrix1.Length;
			int n = matrix2[0].Length;
			double[][] product = new double[m][];
			double sum;
			
			for (int i = 0; i < m; i++) {
				product[i] = new double[n];
				for (int j = 0; j < n; j++) {
					sum = 0.0;
					for (int k = 0; k < matrix1[0].Length; k++) {
						sum += matrix1[i][k] * matrix2[k][j];
					}
					product[i][j] = sum;
				}
			}
			
			return product;
		}
		
		public static Matrix Multiply(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Multiply(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix Multiply(double[][] matrix1, Matrix matrix2)
		{
			return new Matrix(Multiply(matrix1, matrix2.items), false);
		}
		
		public static Matrix Multiply(Matrix matrix1, double[][] matrix2)
		{
			return new Matrix(Multiply(matrix1.items, matrix2), false);
		}
		
		public Matrix Scale(double factor)
		{
			double[][] result = new double[items.Length][];
			
			for (int i = 0; i < result.Length; i++) {
				result[i] = new double[items[0].Length];
				for (int j = 0; j < result[i].Length; j++) {
					result[i][j] = factor * items[i][j];
				}
			}
			
			return new Matrix(result, false);
		}
		
		protected static Matrix Solve(double[][] left, double[][] right)
		{
			double[][] augmentedArray = CreateAugmentedArray(left, right);
			Matrix solution = new Matrix(left.Length, right[0].Length);
			int startColumn = left[0].Length;
			
			GaussJordan(ref augmentedArray);
			
			for (int i = 0; i < solution.RowCount; i++) {
				for (int j = 0; j < solution.ColumnCount; j++) {
					solution.items[i][j] = augmentedArray[i][startColumn + j];
				}
			}
			
			return solution;
		}
		
		public static Matrix Solve(Matrix leftSide, Matrix rightSide)
		{
			return Solve(leftSide.items, rightSide.items);
		}
		
		public Matrix Square()
		{
			int m = items.Length;
			int n = items[0].Length;
			double[][] result = new double[m][];
			for (int i = 0; i < m; ++i) {
				result[i] = new double[m];
				result[i][i] = 0.0;
				for (int k = 0; k < n; ++k) {
					result[i][i] += items[i][k] * items[i][k];
				}
				for (int j = 0; j < i; ++j) {
					result[i][j] = 0.0;
					for (int k = 0; k < n; ++k) {
						result[i][j] += items[i][k] * items[j][k];
					}
					result[j][i] = result[i][j];
				}
			}
			return new Matrix(result, false);
		}
		
		public static double[][] Subtract(double[][] matrix1, double[][] matrix2)
		{
			Exception exception;
			int rows;
			int columns;
			double[][] result;
			
			if (CheckAdditiveCompatibility(matrix1, matrix2, out rows, out columns, out exception)) {
				result = new double[rows][];
				
				for (int i = 0; i < rows; i++) {
					result[i] = new double[columns];
					for (int j = 0; j < columns; j++) {
						result[i][j] = matrix1[i][j] - matrix2[i][j];
					}
				}
				
				return result;
			} else {
				throw exception;
			}
		}
		
		public static Matrix Subtract(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Subtract(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix Subtract(double[][] matrix1, Matrix matrix2)
		{
			return new Matrix(Subtract(matrix1, matrix2.items), false);
		}
		
		public static Matrix Subtract(Matrix matrix1, double[][] matrix2)
		{
			return new Matrix(Subtract(matrix1.items, matrix2), false);
		}
		
		protected static void SwapRows(ref double[][] matrix, int row1, int row2)
		{
			double[] temporary = matrix[row1];
			
			matrix[row1] = matrix[row2];
			matrix[row2] = temporary;
		}
		
		public string ToString(string format)
		{
			string text = String.Empty;
			
			for (int i = 0; i < items.Length; i++) {
				for (int j = 0; j < items[0].Length; j++) {
					if (j > 0) text += "\t";
					text += items[i][j].ToString(format);
				}
				if (i < items.Length - 1) text += Environment.NewLine;
			}
			
			return text;
		}
		
		public override string ToString()
		{
			return ToString("f2");
		}
		#endregion
		
		#region OPERATORS:
		public static implicit operator Matrix(double[][] array)
		{
			return new Matrix(array, false);
		}
		
		public static implicit operator double[][](Matrix matrix)
		{
			return matrix.items;
		}
		
		public static Matrix operator *(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Multiply(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix operator *(double factor, Matrix matrix)
		{
			double[][] items = new double[matrix.RowCount][];
			
			for (int i = 0; i < items.Length; i++) {
				items[i] = new double[matrix.ColumnCount];
				for (int j = 0; j < items[i].Length; j++) {
					items[i][j] = factor * matrix[i, j];
				}
			}
			
			return new Matrix(items, false);
		}
		
		public static Matrix operator *(Matrix matrix, double factor)
		{
			return factor * matrix;
		}
		
		public static Matrix operator +(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Add(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix operator +(double[][] matrix1, Matrix matrix2)
		{
			return new Matrix(Add(matrix1, matrix2.items), false);
		}
		
		public static Matrix operator +(Matrix matrix1, double[][] matrix2)
		{
			return new Matrix(Add(matrix1.items, matrix2), false);
		}
		
		public static Matrix operator -(Matrix matrix1, Matrix matrix2)
		{
			return new Matrix(Subtract(matrix1.items, matrix2.items), false);
		}
		
		public static Matrix operator -(double[][] matrix1, Matrix matrix2)
		{
			return new Matrix(Subtract(matrix1, matrix2.items), false);
		}
		
		public static Matrix operator -(Matrix matrix1, double[][] matrix2)
		{
			return new Matrix(Subtract(matrix1.items, matrix2), false);
		}
		#endregion
	}
	// TODO: Implement determinant, inverse.
}
