﻿/*
 * Készítette a SharpDevelop.
 * Felhasználó: Ouroboros
 * Dátum: 2010.12.29.
 * Idő: 17:04
 * 
 * A sablon megváltoztatásához használja az Eszközök | Beállítások | Kódolás | Szabvány Fejlécek Szerkesztését.
 */
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
//
using TetheredSun.Mathematics;

namespace TetheredSun.Medical
{
	public class IntervalSequence
	{
		protected double[] peaks;
		protected double[] intervals;
		protected double[] successiveDifferences;
		protected List<SequenceParameter> parameters;
		protected string description;
		protected string unit;
		protected IntervalMapping mapping = IntervalMapping.FromPrevious;
		
		
		public IntervalSequence(IList<int> peakIndices, double dt)
		{
			GetPeaks(peakIndices, dt);
			GetIntervals();
			parameters = SequenceParameter.StandardParameters;
		}
		
		public string Description {
			get { return description; }
			set { description = value; }
		}
		
		public double[] Peaks {
			get { return peaks; }
		}
		
		public double[] Intervals {
			get { return intervals; }
		}
		
		public IntervalMapping Mapping {
			get { return mapping; }
			set {
				mapping = value;
				GetIntervals();
			}
		}
		
		public double[] SuccessiveDifferences {
			get { return successiveDifferences; }
		}
		
		public List<SequenceParameter> Parameters {
			get { return parameters; }
		}
		
		public string Unit {
			get { return unit; }
			set { unit = value; }
		}
		
		
		public void EvaluateParameters()
		{
			if (parameters == null) return;
			
			successiveDifferences = null;
			
			foreach (SequenceParameter parameter in parameters) {
				if (parameter is SuccessiveDifferenceParameter) {
					if (successiveDifferences == null) {
						(parameter as SuccessiveDifferenceParameter).Evaluate(intervals, out successiveDifferences);
					} else {
						parameter.Evaluate(successiveDifferences);
					}
				} else {
					parameter.Evaluate(intervals);
				}
			}
		}
		
		protected void GetIntervals()
		{
			if ((peaks == null) || (peaks.Length < 2)) {
				intervals = null;
				return;
			}
			
			intervals = new double[peaks.Length - 1];
			
	
			for (int i = 0; i < intervals.Length; i++) {
				intervals[i] = peaks[i + 1] - peaks[i];
			}
		}
		
		protected void GetPeaks(IList<int> peakIndices, double dt)
		{
			if (peakIndices == null) {
				peaks = null;
				return;
			}
			
			peaks = new double[peakIndices.Count];
			
			for (int i = 0; i < peaks.Length; i++) {
				peaks[i] = peakIndices[i] * dt;
			}
		}
		
		public void SaveAsCsv(string path, string separator, string format)
		{
			if (peaks == null) {
				throw new NullReferenceException("The array containing the peaks is null.");
			}
			if (intervals == null) {
				throw new NullReferenceException("The array containing the intervals is null.");
			}
			if (peaks.Length < 2) {
				throw new InvalidOperationException("Not enough peaks.");
			}
			if (intervals.Length < 1) {
				throw new InvalidOperationException("The array containing the intervals is empty.");
			}
			
			string line = String.Format("[Settings]{0}Description{1}{2}{0}[Intervals]{0}Peak location [{3}]{1}Interval [{3}]",
			                            Environment.NewLine, separator, description, unit);
			int shift = (mapping == IntervalMapping.ToNext) ? 1 : 0;
			
			using (StreamWriter writer = new StreamWriter(path, false, Encoding.UTF8)) {
				writer.WriteLine(line);
				for (int i = 0; i < intervals.Length; i++) {
					line = String.Format("{0}{1}{2}", peaks[i + shift].ToString(format), separator, intervals[i].ToString(format));
					writer.WriteLine(line);
				}
				
				if (parameters != null) {
					line = String.Format("[Parameters]");
					writer.WriteLine(line);
					
					foreach (SequenceParameter parameter in parameters) {
						writer.WriteLine(parameter.ToCsvString(format, separator));
					}
				}
			}
		}
		
	}
}
